#!/bin/env python3
'''
clean wiggle files
'''

import sys;
import argparse;
import collections;
import bisect;

parser=argparse.ArgumentParser(description='Clean wiggle files');

parser.add_argument('-c','--chrom-sizes',type=argparse.FileType('r'),help='Chromosome size file');
parser.add_argument('wigglefile',type=argparse.FileType('r'));

args=parser.parse_args();

chromdict={};
providechromsize=False;
if args.chrom_sizes is not None:
  providechromsize=True;
  for line in args.chrom_sizes:
    field=line.strip().split();
    chromdict[field[0]]=int(field[1]);

def processWholeChrData(crecord,c_chr):
  print('Processing '+str(c_chr)+', #='+str(len(crecord)),file=sys.stderr);
  # create a list of keys
  uo_d={};
  for rec in crecord:
    uo_d[rec[1]]=0;
    uo_d[rec[0]]=0;
  # create an ordered dict
  ## o_d=collections.OrderedDict(sorted(uo_d.items()));
  o_d=uo_d;
  o_d_keys=sorted(o_d.keys());
  # update 
  for rec in crecord:
    startind=bisect.bisect_left(o_d_keys,rec[0]);
    endind=bisect.bisect_left(o_d_keys,rec[1]);
    for ki in o_d_keys[startind:endind]:
      o_d[ki]+=rec[2];
  # output data
  for ki in range(len(o_d_keys)-1):
    print('\t'.join([c_chr, str(o_d_keys[ki]),str(o_d_keys[ki+1]), str(o_d[o_d_keys[ki]])]));


currentchr="";
currentrecord=[];
for line in args.wigglefile:
  field=line.strip().split();
  try:
    chrname=field[0];
    chrstart=int(field[1]);
    chrend=int(field[2]);
    chrval=int(field[3]);
    if providechromsize:
      if chrname not in chromdict.keys():
        continue;
      if chrend > chromdict[chrname]:
        continue;
    if chrname != currentchr:
      processWholeChrData(currentrecord,currentchr);
      currentrecord=[];
      currentchr=chrname;
    currentrecord.append([chrstart,chrend,chrval]);
  except:
    print(line,end='');
    continue;

if len(currentrecord)>0:
  processWholeChrData(currentrecord,currentchr);


