#include <cmath>
#include <algorithm>
#include "graphalgo.h"



/* get the connectivity graph based on the pre-selected parents */
void getcongraph(MInst& inst, mi& allparent, mi& allconn){
  int nsegs=inst.getNSegs();
  //first, set up connectivity graph based on allparents/allchildren
  allconn=mi(nsegs,vi(nsegs,0));
  for(int i=0;i<nsegs;i++)allconn[i][i]=1;
  for(int jj=nsegs-1;jj>=0;jj--){
    for(int ii=0;ii<allparent[jj].size();ii++){
      int cp=allparent[jj][ii];
      if(allconn[cp][jj]==0){
        //copy the connectivity contents
        allconn[cp][jj]=1;
        for(int k=jj+1;k<allconn[jj].size();k++){
          if(allconn[jj][k]==1)allconn[cp][k]=1;
        }
      }
    }
  }
}

/* Get the adjusted children list: remove possible intron childrens. */
vi adjustchildren(mi& allchildren, int tk, MInst& inst,  EnumParameter para,mi& rcount){
  vi ach;
  //remove possible introns from tk
  if(allchildren[tk].size()>1 && tk<allchildren.size()-1){
    //check if tk+1 is in the list?
    for(int i=0;i<allchildren[tk].size();i++){ 
      if(allchildren[tk][i]==tk+1){
        if(para.JS_DISTANT_EXON_ONLY || allchildren[tk+1].size()==0)
          continue;
      }
      ach.push_back(allchildren[tk][i]);
    }
  }else{
    ach=allchildren[tk];
  }
  sort(ach.begin(),ach.end());
  //remove possible introns from the farther points
  vi ach0(ach.size(),1);
  for(int i=0;i<(int)ach.size()-1;i++){
    if(ach[i]==ach[i+1]-1 
      //&& inst.getSegLen(ach[i])<para.JS_MIN_SEG_LEN_TO_CONSIDER
    ){
      //choose the one with more links?
      if(rcount[tk][ach[i+1]]>rcount[tk][ach[i]])
        ach0[i]=0;//
      else
        ach0[i+1]=0;
    }else{
    }
  }
  vi ach2;
  for(int i=0;i<ach.size();i++) if(ach0[i]==1) ach2.push_back(ach[i]);
  return ach2;
}

/* Get the adjusted parent list: remove possible intron parents. */
vi adjustparent(mi& allparent, int tk, MInst& inst, EnumParameter para, mi& rcount){
  vi ach;
  if(allparent[tk].size()>1 && tk>0){
    //check if tk+1 is in the list?
    for(int i=0;i<allparent[tk].size();i++){ 
      if(allparent[tk][i]==tk-1 ){
        if(para.JS_DISTANT_EXON_ONLY|| allparent[tk-1].size()==0) 
          continue;
      }
      ach.push_back(allparent[tk][i]);
    }
  }else{
    ach=allparent[tk];
  }
  //remove possible introns from the farther points
  sort(ach.begin(),ach.end());
  vi ach0(ach.size(),1);
  for(int i=0;i<(int)ach.size()-1;i++){
    if(ach[i]==ach[i+1]-1 
      //&& inst.getSegLen(ach[i+1])<para.JS_MIN_SEG_LEN_TO_CONSIDER
    ){
      if(rcount[ach[i+1]][tk]<rcount[ach[i]][tk])
        ach0[i+1]=0;//
      else
        ach0[i]=0;
    }else{
    }
  }
  vi ach2;
  for(int i=0;i<ach.size();i++) if(ach0[i]==1) ach2.push_back(ach[i]);
  return ach2;
}



/* For each node ns with more than one incoming edges, get the corresponding parents for each edge */
int getparentchain(MInst& inst, mi& allchildren, mi& allparent, int ns, mi& res,EnumParameter para,mi& rcount){
  res.clear();
  vi nspar=adjustparent(allparent,ns,inst,para,rcount);
  for(int i=0;i<nspar.size();i++){
    int tk=nspar[i];
    vi parentchain;
    vi ach=adjustchildren(allchildren,tk,inst,para,rcount);
    vi apa=adjustparent(allparent,tk,inst,para,rcount);
    if(ach.size()==1){//has only 1 children
      parentchain.push_back(tk);
      if(apa.size()==1)
        tk=apa[0];
      else tk=-1;
    }
    while(tk>0&& adjustparent(allparent,tk,inst,para,rcount).size()==1 
      && adjustchildren(allchildren,tk,inst,para,rcount).size()==1){
      parentchain.push_back(tk);
      tk=adjustparent(allparent,tk,inst,para,rcount)[0];
    }
    if(parentchain.size()>0) res.push_back(parentchain);
  }
  return 0;
}

/* For each node ns with more than one outgoing edges, get the corresponding children for each edge */
int getchildrenchain(MInst& inst, mi& allchildren, mi& allparent, int ns, mi& res,EnumParameter para, mi& rcount){
  res.clear();
  vi nsch=adjustchildren(allchildren,ns,inst,para,rcount);
  for(int i=0;i<nsch.size();i++){
    int tk=nsch[i];
    vi childrenchain;
    vi ach=adjustchildren(allchildren,tk,inst,para,rcount);
    vi apa=adjustparent(allparent,tk,inst,para,rcount);
    if(apa.size()==1){
      childrenchain.push_back(tk);
      if(ach.size()>0)
        tk=ach[0];
      else tk=-1;
    }
    while(tk>0 && adjustparent(allparent,tk,inst,para,rcount).size()==1 
      && adjustchildren(allchildren,tk,inst,para,rcount).size()==1){
      childrenchain.push_back(tk);
      tk=adjustchildren(allchildren,tk,inst,para,rcount)[0];
    }
    if(childrenchain.size()>0) res.push_back(childrenchain);
  }
  return 0;
}


/* For each instance, return pairs of segment branches that are subject to correlate. 
  The pair of segment branches are stored in pcheck, each vector contains two numbers, a and b.
  You need to check between all parent chains of node a and all children chains of node b.
  Use getchildrenchain() and getparentchain() to get corresponding segment numbers.
*/
int getcheckingsegments(MInst& inst, mi& rcount, 
  mi& allchildren,
  mi& allparent, 
  EnumParameter para, 
  mi & pcheck){
  
  int nsegs=inst.getNSegs();
  //connectivity graph
  mi allconn;
  getcongraph(inst,allparent,allconn);

  //calculate incoming and outgoing edges for each node
  //however, exclude some "orphan" branches which may be caused by introns.
  //"orphan" parents of x is the node (x-1) which has no incoming edges and only one outgoing edge.
  //similar is "orphan" children
  vi nin(nsegs,0), nout(nsegs,0);
  for(int i=0;i<nsegs;i++){
    nout[i]=adjustchildren(allchildren,i,inst,para,rcount).size();
    nin[i]=adjustparent(allparent,i,inst,para,rcount).size();
  }


  //for each node, if it has more than one incoming/outgoing edges, get its groups of parents/children

  //now check for nodes with more than 1 outgoing edges
  for(int pni=0;pni<nsegs;pni++){
    if(nout[pni]>1){
      //check all its parents to see if there are more than one incoming edges?
      //if yes, this would be a branch
      for(int pparent=0;pparent<=pni;pparent++){
        if(allconn[pparent][pni]==1 && nin[pparent]>1){
           //now, we need to check parents of pparent and childrens of pni to set up pairwise comparison segments
           vi tocheck(2,0);
           tocheck[1]=pni; tocheck[0]=pparent;
           pcheck.push_back(tocheck);
        }//end if
      }//end for(pparent)
    }
  }//end for(pni)
  
  return 0;
}

/* Get the smallest child that appears in all branches from c.
If such child (x) is not found, return -1. 
We also require that c appears in all paths that point to x.*/
int getsmallestcommonchild(mi& rcount, mi&allchildren, mi& allparent, mi& allconn, int c){
  for(int x=c+1;x<allconn[c].size();x++){
    if(allconn[c][x]==1){
      //one of possible children of c
      //check if x is connected by all of c's children?
      bool isnotchild=false;
      for(int j=0;j<allchildren[c].size();j++){
        int nch=allchildren[c][j];
        if(x<nch){
         isnotchild=true;
         break;
        }
        if(x>nch && allconn[nch][x]!=1){
         isnotchild=true;
         break;
        }
      }//end for
      if(isnotchild==false) {
       //also check if x has other ancestors that are not connected to c?
       bool hasdisconnectedancestors=false;
       for(int j=0;j<allconn[x].size();j++){
         if(allconn[j][x]==1 && j!=c){
           if(j<c){
             //j<c
             //check if j connects to x directly without c?
             //then no such common ancestors: all childrens are affected
             for(int k=0;k<allchildren[j].size();k++)
               if(allchildren[j][k]==x){
                 hasdisconnectedancestors=true;
                 return -1;
               }
           }else{
             //c<j<x
             if(allconn[j][c]==0 && allconn[c][j]==0){
               hasdisconnectedancestors=true;
               return -1;
             }
           }
         }
       }
       if(hasdisconnectedancestors==false)
         return x;
      }
    }
  }//end for(x)
  return -1;
}




/* Get reference segments: segments that appear in one isoform. */
int getreferencesegments(MInst& inst, mi& rcount, 
  mi& allchildren,
  mi& allparent, 
  EnumParameter para, 
  mi & refs){

  
  int nsegs=inst.getNSegs();
  //incoming and outgoing edges for each node
  vi nin(nsegs,0), nout(nsegs,0);
  for(int i=0;i<nsegs;i++){
    nout[i]=allchildren[i].size();
    nin[i]=allparent[i].size();
  }

  //connectivity graph
  mi allconn;
  getcongraph(inst,allparent,allconn);

  vi included(nsegs,0);
  for(int i=0;i<allchildren.size();i++){
    if(included[i]==0  ){
      vi thischain;
      thischain.push_back(i);
      included[i]=1;
      //get the smallest children that appear in all its branches
      int tsci=getsmallestcommonchild(rcount,allchildren,allparent,allconn,i);
      while(tsci!=-1){
        included[tsci]=1;
        thischain.push_back(tsci);
        tsci=getsmallestcommonchild(rcount,allchildren,allparent,allconn,tsci);
      }
      if(thischain.size()>1) refs.push_back(thischain);
    }
  }
  
  return 0;
}

/* Check if two segments, a and b, are coexpressed in annotation (inst.refs)
Return value: -1 if error, 0 if not, 1 if true. */
int iscoexpressioninanno(MInst& inst, int a, int b){
  if(a<0 || a>=inst.getNSegs())return -1;
  if(b<0 || b>=inst.getNSegs())return -1;

  int nconsistent=0;
  for(int i=0;i<inst.refs.size();i++){
    if(inst.refs[i][a]==inst.refs[i][b])nconsistent++;
  }
  if(nconsistent==inst.refs.size())return 1;
  else return 0;
}

/* Check if two segments, a and b, are exclusive in annotation (inst.refs)
Return value: -1 if error, 0 if not, 1 if true. */
int isexclusiveinanno(MInst& inst, int a, int b){
  if(a<0 || a>=inst.getNSegs())return -1;
  if(b<0 || b>=inst.getNSegs())return -1;

  int nconsistent=0;
  for(int i=0;i<inst.refs.size();i++){
    if(inst.refs[i][a]!=1 && inst.refs[i][b]!=1)nconsistent++;
  }
  if(nconsistent==inst.refs.size())return 1;
  else return 0;
}
/* Get the mean value of a set of vectors */
double getmean(vd& vals){
  double dl=0;
  if(vals.size()==0)return 0;
  for(int i=0;i<vals.size();i++) dl+=vals[i];
  return dl/(int)vals.size();
}

/* get the std value of a set of vectors */
double getstd(vd& vals){
  if(vals.size()<=1)return 0;
  double mv=getmean(vals);
  double dl=0;
  for(int i=0;i<vals.size();i++) dl+= (vals[i]-mv)*(vals[i]-mv);
  dl/=(int)vals.size();
  return sqrt(dl);
}
