/*
 Processsam
 Update history
  09/23/2011: add support of CIGAR character "D" and "I"
  09/25/2011: fix STDIN bug
  10/18/2011: change STDIN to -.
*/


#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <vector>
#include <algorithm>
#include <climits>
#include "instance.h"
#include "structdef.h"


int readOldInstance(string insFile, vector<OldInstance>& result, int nInstance){
  ifstream ifs(insFile.c_str());
  int instCount=0;
  if(!ifs.is_open()){
    cerr<<"Error inst "<<instCount+1<<" opening files!\n";
    return -1;
  }
  OldInstance ins;


  while(!ifs.eof()){//main loop
    string oneline;
    //1st line
    getline(ifs,oneline);
    if(oneline.size()<2)break;
    stringstream ss(oneline);
    
    string tmp;
    int id;
    ss>>tmp>>id;
    if(tmp!="Instance"){
      cerr<<"Error inst "<<instCount+1<<": should expect Instance here.\n";
    }
    ins.id=id;
    //2nd line
    getline(ifs,oneline);
    stringstream ss2(oneline);
    int nGenes;

    ss2>>tmp>>nGenes;
    ins.nIsoforms=nGenes;
    if(tmp!="genes"){
      cerr<<"Error inst "<<instCount+1<<": should expect genes here.\n";
    }
    //3rd line
    string chromosome;
    char strand;
    ins.isoformNames=vector<string>(nGenes);
    ins.isoformStrands=vector<char>(nGenes);
    ins.isoformStarts=vector<vector<int> >(nGenes);
    ins.isoformEnds=vector<vector<int> >(nGenes);
    for(int i=0;i<nGenes;i++){
      getline(ifs,oneline);
      stringstream ss3(oneline);
      string isoName;
      ss3>>isoName;//name
      ss3>>chromosome;//chromosome
      ss3>>strand;//strand
      ins.pos=chromosome;
      ins.isoformNames[i]=(isoName);
      ins.isoformStrands[i]=(strand);

      long tmp1,tmp3;
      ss3>>tmp1>>tmp3;
      //read , separated start and end positions
      int situ=0;//0: reading start; 1: read end; 2: finish
      //int nstartrec=0,nendrec=0;
      do{
        ss3>>tmp1;
        if(situ==0)
          ins.isoformStarts[i].push_back(tmp1);
        else
          ins.isoformEnds[i].push_back(tmp1);
        char tmpc;
        ss3>>tmpc;
        if(tmpc!=','){
          ss3.unget();
          situ++;
        }
        else{
        }
        
        if(ss3.eof())break;
      }while(true);
      
    }
    //segs line
    getline(ifs,oneline);
    stringstream ss4(oneline);
    int nExons;
    ss4>>tmp>>nExons;
    if(tmp!="segs"){
      cerr<<"Error inst "<<instCount+1<<": should expect segs here.\n";
    }
    ins.nExons=nExons;
    ins.exonStrands=vector<char>(nExons);
    ins.exonStarts=vector<long>(nExons);
    ins.exonEnds=vector<long>(nExons);
    ins.exonFlags=vector<int>(2*nExons);
    for(int i=0;i<nExons;i++){//read segs line
      getline(ifs,oneline);
      stringstream ss5(oneline);
      char exstrc;
      ss5>>tmp>>exstrc;
      if(tmp!=chromosome){
        cerr<<"Error inst "<<instCount+1<<": segs chromosome should be consistent with instance chromosome.\n";
      }
      ins.exonStrands[i]=(exstrc);
      long eSt, eEd;
      ss5>>eSt>>eEd;
      ins.exonStarts[i]=(eSt);
      ins.exonEnds[i]=(eEd);
      int zeroone1, zeroone2;
      ss5>>zeroone1>>zeroone2;
      ins.exonFlags[2*i]=(zeroone1);
      ins.exonFlags[2*i+1]=(zeroone2);
    }
    
    //exon length
    getline(ifs,oneline);
    stringstream ss6(oneline);
    ss6>>tmp;
    if(tmp!="len"){
      cerr<<"Error inst "<<instCount+1<<": should expect len here.\n";
    }
    ins.exonLens=vector<int>(nExons);
    for(int i=0;i<nExons;i++)
    {
      int exoL;
      ss6>>exoL;
      ins.exonLens[i]=(exoL);
    }
    //isoforms line
    getline(ifs,oneline);
    stringstream ss7(oneline);
    int ni1,ni2;
    ss7>>tmp>>ni1>>ni2;
    if(tmp!="isoforms"){
      cerr<<"Error inst "<<instCount+1<<": should expect isoforms here.\n";
    }
    if(ni1!=ni2 || ni1!=nGenes){
      cerr<<"Error inst "<<instCount+1<<": not correct formats.\n";
    }
    ins.isoforms=vector<vector<int> > (nGenes, vector<int>(nExons));
    ins.expLevels=vector<double>(nGenes);
    for(int i=0;i<nGenes;i++){
      getline(ifs,oneline);
      stringstream ss8(oneline);
      int tki;
      ss8>>tmp>>tki;
      if(tmp!="K" || tki!=i){
        cerr<<"Error inst "<<instCount+1<<": not correct formats in isoform matrix.\n";
      }
      for(int j=0;j<nExons;j++){
        int jie;
        ss8>>jie;
        if(jie!=0&&jie!=1){
          cerr<<"Error inst "<<instCount+1<<": isoform matrix must be 0/1.\n";
        }
        ins.isoforms[i][j]=(jie);
      }
      double explv;
      ss8>>explv;
      ins.expLevels[i]=(explv);
      
    }
    //SE segs line
    getline(ifs,oneline);
    stringstream ss9(oneline);
    int nSegs;
    ss9>>tmp>>nSegs;
    if(tmp!="SESegs"){
      if(tmp=="Reads"){//optional session for Reads
        int nReads=nSegs;
        //parse reads
        ins.readPattern=vector<vector<int> >(nReads, vector<int>(4,-1));
        ins.readCount=vector<int>(nReads,0);
        for(int i=0;i<nReads;i++){
          getline(ifs,oneline);
          stringstream ssr(oneline);
          int eid;
          ssr>>eid;
          ins.readPattern[i][0]=eid;
          char tc;
          ssr>>tc;
          if(tc!=';'){
            ssr.unget();
            ssr>>eid;
            ins.readPattern[i][1]=eid;
            ssr>>tc;
            if(tc!=';'){
              //Warning here: does not support mapping information of more than 2 exons!
              //cerr<<"Error format: should expect another segs here for instance "<<ins.id<<"\n";
              ins.readCount[i]=0;
              ins.readPattern[i][0]=ins.readPattern[i][1]=-1;
              continue;
            }
          }
          else{
            ins.readPattern[i][1]=ins.readPattern[i][0];
          }
          int ntp;
          ssr>>ntp;
          int countsemicolumn=count(oneline.begin(),oneline.end(),';');
          if(countsemicolumn==1){
            //single end pattern: ntp should be the read count
            ins.readCount[i]=ntp;
            continue;
          }
          else if(countsemicolumn==2){
            eid=ntp;
            ins.readPattern[i][2]=eid;
            char tc;
            ssr>>tc;
            if(tc!=';'){
              ssr.unget();
              ssr>>eid;
              ins.readPattern[i][3]=eid;
              ssr>>tc;
              if(tc!=';'){
                //Warning here: does not support mapping information of more than 2 exons!
                //cerr<<"Error format: should expect another segs here for instance "<<ins.id<<"...\n";
                ins.readCount[i]=0;
                ins.readPattern[i][0]=ins.readPattern[i][1]=-1;
              }
            }
            else{
              ins.readPattern[i][3]=ins.readPattern[i][2];
            }
            ssr>>ntp;
            ins.readCount[i]=ntp;

          }//end if semicolumn count
          else{
            cerr<<"Error: incorrect number of semicolumn counts:"<<countsemicolumn<<endl;
          }
          

        }
        //read SESegs here
        getline(ifs,oneline);
        stringstream ssx(oneline);
        ssx>>tmp>>nSegs;
      }
      else{
        cerr<<"Error inst "<<instCount+1<<": should expect SESegs or Reads here.\n";
      }
    }
    ins.seSegs=vector<vector<int> >(nSegs, vector<int>(2));
    for(int i=0;i<nSegs;i++){
      getline(ifs,oneline);
      stringstream ss10(oneline);
      int segst,seged;
      ss10>>segst>>seged;
      ins.seSegs[i][0]=(segst);
      ins.seSegs[i][1]=(seged);
      
    }//end of nSegs loop
    
    result.push_back(ins);

    instCount++;
    //if(instCount%100==1)
      //cout<<"reading instance "<<instCount<<endl;
    if(nInstance>=0 && instCount>nInstance)break;
  }//end of ifs loop 



  ifs.close();
  return 0;
}

void printOldInstance(ostream& out, OldInstance& ins){
  out<<"Instance\t"<<ins.id<<endl;
  out<<"genes\t"<<ins.nIsoforms<<endl;
  for(int i=0;i<ins.nIsoforms;i++){
    out<<ins.isoformNames[i]<<"\t"<<ins.pos<<"\t"
      <<ins.isoformStrands[i]<<"\t"
      <<ins.isoformStarts[i][0]<<"\t"<<ins.isoformEnds[i].back()<<"\t";
    for(int j=0;j<ins.isoformStarts[i].size();j++){
      out<<ins.isoformStarts[i][j];
      if(j==ins.isoformStarts[i].size()-1)
        out<<"\t";
      else
        out<<",";
    }
    for(int j=0;j<ins.isoformEnds[i].size();j++){
      out<<ins.isoformEnds[i][j];
      if(j==ins.isoformEnds[i].size()-1)
        out<<"\n";
      else
        out<<",";
    }
  }
  out<<"segs\t"<<ins.nExons<<endl;
  int nexonflags=0;
  for(int i=0;i<ins.nExons;i++){
    out<<ins.pos<<"\t"<<ins.exonStrands[i]<<"\t"<<ins.exonStarts[i]<<"\t"<<ins.exonEnds[i]<<"\t";
    out<<ins.exonFlags[2*i]<<"\t"<<ins.exonFlags[2*i+1]<<endl;
  }
  out<<"len\t";
  for(int i=0;i<ins.nExons;i++){
    out<<ins.exonLens[i]<<"\t";
  }
  out<<endl;
  out<<"isoforms\t"<<ins.nIsoforms<<"\t"<<ins.nIsoforms<<endl;
  for(int i=0;i<ins.nIsoforms;i++){
    out<<"K\t"<<i<<"\t";
    for(int j=0;j<ins.nExons;j++)
      out<<ins.isoforms[i][j]<<"\t";
    out<<ins.expLevels[i]<<endl;
  }
  out<<"Reads\t"<<ins.readPattern.size()<<endl;
  for(int i=0;i<ins.readPattern.size();i++){
    for(int j=0;j<ins.readPattern[i].size();j++)
      out<<ins.readPattern[i][j]<<" ";
    out<<"\tCount: "<<ins.readCount[i]<<endl;
  }
  out<<"SESegs\t"<<ins.seSegs.size()<<endl;
  for(int i=0;i<ins.seSegs.size();i++)
    out<<ins.seSegs[i][0]<<"\t"<<ins.seSegs[i][1]<<endl;

  
}

/**
Building instance index
NOTICE: Incomplete
*/
int buildOldInstanceIndex(vector<OldInstance>& result){
  
  map<string,int>chromIndex;
  map<pair<int,long>, int> instanceIndex;
  
  //first scan whole instance to get a list of possible chromosomes
  for(int i=0;i<result.size();i++){
    if(chromIndex.count(result[i].pos)==0){
      chromIndex[result[i].pos]=0;
    }
  }
  //modify the list to change 0 to corresponding id
  int chrindex=0;
  for(map<string,int>::iterator itp=chromIndex.begin();itp!=chromIndex.end();itp++){
    itp->second=chrindex;
    chrindex++;
  }
  for(map<string,int>::iterator itp=chromIndex.begin();itp!=chromIndex.end();itp++){
    cout<<itp->first<<":"<<itp->second<<endl;
  }
  
  for(int i=0;i<result.size();i++){
    pair<int,long> querykey=make_pair<int,long>(chromIndex[result[i].pos], result[i].exonStarts[0]);
    instanceIndex[querykey]=i;
  }
  return 0;
}

//------------------------------------------------------------------
 
/*
Read a matrix like file and save the items into a vector of vector
This function reads a "block" data: the 1st line it encounters should indicate two items:
segname	nlines
indicating the name of the segment and the number of lines following the first line

For the data line: 
  lines beginning with '#' will be considered as comment line;
  If lines begin with '!', the first field will be ignored.
*/
template<class T>
int fillMat(vector<vector<T> > &m,istream & ifs,string segname){
  //ifstream ifs(filename);
  string oneline;
  getline(ifs,oneline);
  stringstream ss0(oneline);
  string tsegname;
  int nline=0;
  ss0>>tsegname>>nline;
  if(tsegname!=segname){
    cerr<<"Error: inconsistent block name. Expect "<<segname<<", encounter "<<tsegname<<endl;
  }
  int n=0;
  while(true){
    if(n>=nline)break;
    getline(ifs,oneline);
    if(ifs.eof())break;
    if(oneline.size()>0 && oneline[0]=='#')continue;
    n++;
    vector<T> ct;
    stringstream ss(oneline);
    if(oneline[0]=='!'){
      string tmpstr;
      ss>>tmpstr;
    }
    while(!ss.eof()){
      T tp;
      ss>>tp;
      if(!ss.fail())
        ct.push_back(tp);
      else
        break;
    }
    m.push_back(ct);
  }
  //ifs.close();
  return 0;
}

template<class T>
int readPair(string oneline, string key, T& value){
  stringstream ss(oneline);
  string kk;
  ss>>kk>>value;
  if(kk!=key){
    return -1;
  }
  return 0;
}


int Instance::readFromStream(istream& ifs){
  string oneline;
  do{
    getline(ifs,oneline);
  }while(!ifs.eof() && oneline.size()>0 && oneline[0]!='I');
  if(ifs.eof()|| oneline.size()==0)return -1;
  
  if(readPair<string>(oneline,"Instance",id)==-1){
    cerr<<"Error: expect Instance here.\n";
    return -1;
  }
  cout<<"Reading instance "<<id<<"...\n";
  getline(ifs,oneline);
  stringstream ss2(oneline);
  string sbound;
  ss2>>sbound>>chr>>boundary.first>>boundary.second>>dir;
  if(sbound!="Boundary"){
    cerr<<"Error: expect Boundary here.\n";
    return -1;
  }
  getline(ifs,oneline);
  if(readPair<int>(oneline,"ReadLen",readlen)==-1){
    cerr<<"Error: expect ReadLen here.\n";
    return -1;
  }
  if(fillMat<double>(segs,ifs,"Segs")==-1){
    cerr<<"Error reading segments.\n";
    return -1;
  }
  if(fillMat<int>(refs,ifs,"Refs")==-1){
    cerr<<"Error reading Refs.\n";
    return -1;
  }
  getline(ifs,oneline);
  if(readPair<long>(oneline,"Reads",nreads)==-1){
    cerr<<"Error: expect Reads here.\n";
    return -1;
  }
  if(fillMat<int>(sgtypes,ifs,"SGTypes")==-1){
    cerr<<"Error reading SGTypes.\n";
    return -1;
  }
  //process sgtypes
  for(int i=0;i<sgtypes.size();i++){
    sgdirs.push_back(sgtypes[i].back());//direction
    sgtypes[i].pop_back();
    sgcounts.push_back(sgtypes[i].back());//counts
    sgtypes[i].pop_back();
  }

  //Process PETypes
  getline(ifs,oneline);
  stringstream ss3(oneline);
  string petypestr;
  int nperead,npetype;
  ss3>>petypestr>>nperead>>npetype;
  if(petypestr!="PETypes"){
    cerr<<"Error: expect PETypes.\n";
    return -1;
  }
  for(int np=0;np<npetype;np++){
    getline(ifs,oneline);
    stringstream ss4(oneline);
    //get type and type read, save to petype and petypecount
    int type1,type2,ntp;
    ss4>>type1>>type2>>ntp;
    vi tpt; tpt.push_back(type1);tpt.push_back(type2);
    petypes.push_back(tpt);
    pecounts.push_back(ntp);
    //get paired-end distance, save to petypedist
    getline(ifs,oneline);
    stringstream ss5(oneline);
    vi pepd(ntp);
    vi pepdcnt(ntp);
    for(int i=0;i<ntp;i++){
      int tmpdt,tmpdtc;char c;
      ss5>>tmpdt>>c>>tmpdtc; //expect c=':'
      pepd[i]=tmpdt;
      pepdcnt[i]=tmpdtc;
    }
    pedists.push_back(pepd);
    pedistscount.push_back(pepdcnt);
    
  }
  return 0;
}

/* Check if the boundaries are identical */
bool Instance::checkBoundaryWithOtherInst(Instance& r){
  if(r.chr!=chr)return false;
  if(r.segs.size()!=segs.size())return false;
  //check segments one by one
  for(int i=0;i<segs.size();i++){
    if(segs[i][0]!=r.segs[i][0])return false;
    if(segs[i][1]!=r.segs[i][1])return false;
  }
  return true;
}
/* Convert multiple instances to minst, and send to ostream */
int inst2minst(ostream& out,vector<Instance>& vinst,vl& totalreads){
  if(vinst.size()==0)return -1;
  if(totalreads.size()!=vinst.size())return -1;
  Instance & finst=vinst[0];
  md& segs=finst.segs;
  out<<"Instance\t"<<finst.id<<endl;
  out<<"segment\t"<<int(segs.size())<<endl;
  for(int i=0;i<segs.size();i++){
    out<<"!"<<i<<"\t"<<int(segs[i][0])<<"\t"<<int(segs[i][1])<<endl;
  }
  out<<"segcount\t"<<int(segs.size())<<endl;
  for(int i=0;i<segs.size();i++){
    out<<"!"<<i<<"\t";
    for(int j=0;j<vinst.size();j++){
      out<<int(vinst[j].segs[i][3])<<"\t";
    }
    out<<endl;
  }
  out<<"segcvg\t"<<int(segs.size())<<endl;
  for(int i=0;i<segs.size();i++){
    out<<"!"<<i<<"\t";
    for(int j=0;j<vinst.size();j++){
      out<<(vinst[j].segs[i][8])<<"\t";
    }
    out<<endl;
  }
  out<<"segp0\t"<<int(segs.size())<<endl;
  for(int i=0;i<segs.size();i++){
    out<<"!"<<i<<"\t";
    for(int j=0;j<vinst.size();j++){
      out<<(vinst[j].segs[i][7])<<"\t";
    }
    out<<endl;
  }
  out<<"segmaxcvg\t"<<int(segs.size())<<endl;
  for(int i=0;i<segs.size();i++){
    out<<"!"<<i<<"\t";
    for(int j=0;j<vinst.size();j++){
      out<<(vinst[j].segs[i][4])<<"\t";
    }
    out<<endl;
  }
  //merge sgtypes
  map<vi,vi>alltypes;
  map<vi,int> alltypedirs;
  for(int j=0;j<vinst.size();j++){ 
    for(int k=0;k<vinst[j].sgtypes.size();k++){
      alltypes[vinst[j].sgtypes[k]]=vi(vinst.size());
      alltypedirs[vinst[j].sgtypes[k]]=vinst[j].sgdirs[k];
    }
  }
  for(int j=0;j<vinst.size();j++){ 
    for(int k=0;k<vinst[j].sgtypes.size();k++){
      alltypes[vinst[j].sgtypes[k]][j]=vinst[j].sgcounts[k];
    }
  }
  //output sgtype
  map<vi,vi>::reverse_iterator rmit;
  map<vi,int> typeorders;
  out<<"sgtype\t"<<alltypes.size()<<endl;
  int index=1;
  for(rmit=alltypes.rbegin();rmit!=alltypes.rend();rmit++){
    const vi& fbin=rmit->first;
    out<<"!"<<index-1<<"\t";
    typeorders[fbin]=index;index++;
    for(int j=0;j<fbin.size();j++)
      out<<fbin[j]<<" ";
    out<<alltypedirs[fbin]<<" ";
    out<<endl;
  }
  //sgcount
  out<<"sgcount\t"<<alltypes.size()<<endl;
  int sgindex=0;
  for(rmit=alltypes.rbegin();rmit!=alltypes.rend();rmit++){
    vi& fbin=rmit->second;
    out<<"!"<<sgindex++<<"\t"; 
    for(int j=0;j<fbin.size();j++)
      out<<fbin[j]<<"\t";
    out<<endl;
  }
  map<vi,vi> allpetypes;
  //merge petypes
  //Notice that the sgtypes are now reshuffled
  //convert petypes
  for(int j=0;j<vinst.size();j++){
    for(int k=0;k<vinst[j].petypes.size();k++){
      for(int l=0;l<vinst[j].petypes[k].size();l++){
        int &currenttypeid=vinst[j].petypes[k][l];
        vi& oldtype=vinst[j].sgtypes[currenttypeid-1];
        int newtypeid=-1;
        if(typeorders.count(oldtype)>0)
          newtypeid=typeorders[oldtype];
        if(newtypeid==-1){
          cerr<<"Error: sgtype not recorded.\n";
        }
        currenttypeid=newtypeid;
      }
    }
  }
  for(int j=0;j<vinst.size();j++){ 
    for(int k=0;k<vinst[j].petypes.size();k++){
      allpetypes[vinst[j].petypes[k]]=vi(vinst.size());
    }
  }
   for(int j=0;j<vinst.size();j++){ 
    for(int k=0;k<vinst[j].petypes.size();k++){
      allpetypes[vinst[j].petypes[k]][j]=vinst[j].pecounts[k];
    }
  } 
 //output pecount
  map<vi,vi>::iterator mit;
  out<<"petype\t"<<allpetypes.size()<<endl;
  int petypeindex=0;
  for(mit=allpetypes.begin();mit!=allpetypes.end();mit++){
    const vi& fbin=mit->first;
    out<<"!"<<petypeindex++<<"\t";
    for(int j=0;j<fbin.size();j++)
      out<<fbin[j]<<" ";
    out<<endl;
  }
  //pecount
  out<<"pecount\t"<<allpetypes.size()<<endl;
  int pectindex=0;
  for(mit=allpetypes.begin();mit!=allpetypes.end();mit++){
    vi& fbin=mit->second;
    out<<"!"<<pectindex++<<"\t";
    for(int j=0;j<fbin.size();j++)
      out<<fbin[j]<<"\t";
    out<<endl;
  }
 
  //readlen
  out<<"readlen\t1\n";
  for(int i=0;i<vinst.size();i++) out<<vinst[i].readlen<<"\t";out<<endl;
  //totalreads
  out<<"totalreads\t1\n";
  for(int i=0;i<totalreads.size();i++)out<<totalreads[i]<<"\t"; out<<endl;
  //otherinfo
  out<<"otherinfo\t"<<(int)2+vinst[0].refs.size()<<"\n";
  out<<"Chr\t"<<finst.chr<<endl;
  out<<"Dir\t"<<finst.dir<<endl;
  //References
  for(int i=0;i<vinst[0].refs.size();i++){
    out<<"Refs\t";
    for(int j=0;j<vinst[0].refs[i].size();j++) if(vinst[0].refs[i][j]>0) out<<j<<" ";
    out<<endl;
  }
  
  
  return 0;
}

/* Sum up the Reads field in the Instance file */
long gettotalreads(string instfile){
  ifstream ifs(instfile.c_str());
  if(!ifs.is_open()){
    cerr<<"Error opening file "<<instfile<<endl;
    return -1;
  }
  long total=0;
  while(true){
    string oneline;
    getline(ifs,oneline);
    if(ifs.eof())break;
    if(oneline.size()<3 || oneline[0]!='R' || oneline[1]!='e')continue;
    long nreads=0;
    if(readPair<long>(oneline,"Reads",nreads)==-1){
      continue;
    }
    total+=nreads;
  }
  ifs.close();
  return total;
}
