/*
  Isoform enumeration related functions.
*/
#ifndef MENUMISO_H
#define MENUMISO_H

#include "minst.h"


/*
Parameters for isoform enumeration
*/
struct EnumParameter{
  /* Maximum number of isoforms allowed */
  int maxniso;
  /* Only return the number of isoforms */
  bool numonly;
  /* print debug information */
  bool debug;
  /* Return immediately if the maximum number of isoform exceeds maxniso */
  bool immreturn;

  /* the minimum fraction to the max junction read when enumerating */
  float minjuncfrac;

  /* For intron retention, the fraction should be set up higher: minadjjuncfrac*minjuncfrac is used.
  */
  float minadjjuncfrac;
  /* For intron retention, only the mean coverage is greater than the fraction of the previous one is enumerated. */
  float minintronretentioncvgfrac;
  /* For intron retention, only segment with length greater than this number is considered intron retention */
  int minintronretentionlen;

  /* Parameter of alpha in the objective function */
  double RDSUPPORT_ALPHA;

  /* The maximum number of constraints added for supporting reads. If there are too many constraints, disable the supporting reads. */
  int MAX_SUPP_CONSTRAINT;

  /* Whether the assembly process should be done through sample by sample? */
  bool assemble_by_sample;

  int current_sample;


  /* Parameters for Jensen-Shannon test */

  /* whether JS test is performed? */
  bool DO_CORRELATION;

  /* We collect two distributions for all JS divergense: JS for all pairwise segments and JS for only reference segments.
     The mean of all JS must be larger than mean(refJS)+JS_MIN_DISTRIB_DIST*std(refJS) to allow for further analysis. */
  double JS_MIN_DISTRIB_DIST;

  /* the JS value must be > mean(allJS)+JS_MIN_EX_DIST*std(allJS) to be considered as exclusive expression. */
  double JS_MIN_EX_DIST;

  /* the JS value must be < mean(refJS)+JS_MIN_EX_DIST*std(refJS) to be considered as coexpression. */
  double JS_MIN_CO_DIST;

  /* When determining parents or children, skip adjacent exons entirely. */
  bool JS_DISTANT_EXON_ONLY;
   
  /* When determining segments, the minimum length of the segment */
  int JS_MIN_SEG_LEN_TO_CONSIDER;

  EnumParameter();

};


/*
For a given segment c, select its possible children
*/
void selectchildren(MInst& inst, int c,
  mi& rcount,vi& children,EnumParameter& para);

/* Enumerate isoforms */
int enumiso(MInst& inst, mi & res,int maxniso); 

/* The main interface.
   Enumerate isoforms, and calculate their expression levels 
*/
int enumandcalexp(MInst& inst, mi& res, md& explv);
#endif
