#include <iostream>
#include <fstream>
#include <sstream>
#include "rangeset.h"
#include "bedio.h"
#include "mergejunc.h"

using namespace std;


/* Stores the count information */
AllChrJcount JCOUNT;
/* stores the direction information 
   0: undetermined, >0: junctions starts; <0: junction ends. 
*/
AllChrJdir JDIR;

int NSAMPLE=0;

/* Min number of junctions supporting this position */
int MIN_JC=2;

/* Min number of samples supporting this position */
int MIN_JSAMPLE=2;

void printmsg(){
  cerr<<"Usage: mergejunc {OPTIONS} <BED file 1> <BED file 2> ...\n";
}

/*
Process one bed file.
Only start/end positions are used.
*/
int processonebed(string filename,int id){
  ifstream ifs(filename.c_str());
  if(!ifs.is_open()){
    cerr<<"Error opening BED file "<<filename<<endl;
    return -1;
  }
  int linecount=0;
  string prevchr="";
  range_t currentrange;
  vector<bedrec> rs;
  while(true){
    string oneline;
    getline(ifs,oneline);
    linecount++;
    bedrec rec;
    if(ifs.eof())break;
    if(scanBed(oneline,rec)==-1)continue;
    if(rec.segstart.size()!=2){
      cerr<<"Error: incorrect junction format.\n";
      continue;
    }
    long jstart=rec.segend[0]+1;
    long jend=rec.segstart[1];
    if(JCOUNT[rec.chr].count(jstart)==0) JCOUNT[rec.chr][jstart]=vector<int>(NSAMPLE);
    if(JCOUNT[rec.chr].count(jend)==0) JCOUNT[rec.chr][jend]=vector<int>(NSAMPLE);
    JCOUNT[rec.chr][jstart][id]+=rec.score;
    JCOUNT[rec.chr][jend][id]+=rec.score;
    if(rec.dir=='+'){//positive
      JDIR[rec.chr][jstart]+=rec.score;
      JDIR[rec.chr][jend]-=rec.score;
    }else{//negative
      JDIR[rec.chr][jstart]-=rec.score;
      JDIR[rec.chr][jend]+=rec.score;
    }
    
  }
  ifs.close();
  return 0;
}

void junction2stdout(){
  AllChrJcount:: iterator jci;
  for(jci=JCOUNT.begin();jci!=JCOUNT.end();jci++){
    ChrJcount & mchi=jci->second;
    ChrJcount::iterator pit;
    for(pit=mchi.begin();pit!=mchi.end();pit++){
      vi & data=pit->second;
      int nc=0,nsample=0;
      for(int i=0;i<data.size();i++){
        nc+=data[i];
        nsample+=data[i]>0?1:0;
      }
      if(nc>=MIN_JC || nsample >=MIN_JSAMPLE){
        cout<<jci->first<<"\t"<<pit->first<<"\t"<<pit->first<<"\t";
        int cdir=JDIR[jci->first][pit->first];
        if(cdir>0)  cout<<"+";
        else if(cdir<0) cout<<"-";
        else cout<<".";
          
        for(int i=0;i<data.size();i++){
          cout<<"\t"<<data[i];
        }
        cout<<endl;
      }//end if
    }//end pit for loop
  }//end jci for loop
}

int mergejunc_main(int argc, char* argv[]){
  if(argc==1){
    printmsg();
    return 0;
  }
  //processing arguments
  int opt=1;
  vector<string> fnames;
  while(opt<argc){
    fnames.push_back(string(argv[opt]));
    opt++;
  }
  NSAMPLE=fnames.size();
  if(NSAMPLE<1) return -1;
  for(int i=0;i<fnames.size();i++){
    processonebed(fnames[i],i);
  }
  junction2stdout();
  return 0;
}
