#include <vector>
#include <iostream>
#include "mpredexp.h"
#include "isolassocpp/mathlib.h" // LassoWithConstraint
#include "isolassocpp/emfunction.h" // LassoWithConstraint

/* check if two binary vectors of the same size are overlapping */
bool vecoverlap(vi& x, vi& y){
  for(int i=0;i<x.size();i++) if(x[i]*y[i]>0)return true;
  return false;
}
/* return the union of both binary vectors in x*/
void vecunion(vi&x, vi&y){
  for(int i=0;i<x.size();i++)  x[i]=(x[i]+y[i]>0?1:0);
}

/* group the assembly results. the index are stored in munionindex. */
int groupres(mi& mbin, mi& mindex){
 
  mi munion=mbin;
  mi munionindex;
  vi validmunion(mbin.size(),1);
  //initialize
  for(int i=0;i<mbin.size();i++){
    munionindex.push_back(vi(1,i));
  }
  while(true){
    bool haschanged=false;
    for(int i=0;i<munion.size();i++){
      if(validmunion[i]==0)continue;
      for(int j=i+1;j<munion.size();j++){
        if(validmunion[j]==0)continue;
        if(vecoverlap(munion[i],munion[j])==true){
          vecunion(munion[i],munion[j]);
          validmunion[j]=0;
          for(int k=0;k<munionindex[j].size();k++)
            munionindex[i].push_back(munionindex[j][k]);
          haschanged=true;
        }
      }//end for j
    }//end for i
    if(haschanged==false)break;
  }
  mindex.clear();
  for(int i=0;i<validmunion.size();i++){
    if(validmunion[i]==1) mindex.push_back(munionindex[i]);
  }

  return 0;
}

/*
 Call LassoWithConstraint() function to calculate expression levels
*/
int gsllassoconstraint(MInst& inst,md& xval, vd& yval, vd& pred){
  
  double lambda=1000;//set up a large lambda (unbounded)
  NewInstance ins;
  vector<string> varargin;
  vector<double> w;
  double fval;
  long exitflag;
  //set up NewInstance
  ins.id='0';
  //set up varargin. it's important to set useexoncon and usejunccon off 
  varargin.push_back("--useexoncon");varargin.push_back("false");
  varargin.push_back("--usejunccon");varargin.push_back("false");
  //lasso with constraint
  LassoWithConstraint(xval, yval,lambda,ins,varargin, w, fval, exitflag);
  cout<<"fval:"<<fval<<",flag:"<<exitflag<<endl;
  cout<<"w.size()="<<w.size()<<",pred.size()="<<pred.size()<<endl;
  pred.assign(w.begin(),w.end());
  if(pred.size()<xval[0].size()){
    cerr<<"Warning: unsucessful quadprog.\n";
    pred.resize(xval[0].size());
  }
  
  return 0;
}

/*
Using GSL library to calculate (XX'+\lambda)^-1*Xy
Not tested
*/
int  gslmatinv(md& xval, vd& yval, vd& pred){
//  double lambda=1e-5;
//  gsl_matrix *result = gsl_matrix_alloc(xval.size(), xval.size());
//  gsl_matrix *inverse = gsl_matrix_alloc(xval.size(), xval.size());
//  for (long i = 0; i < xval.size(); ++i)
//    for (long j = 0; j < xval.size(); ++j){
//      double tmp = 0;
//      for (long k = 0; k < xval[i].size(); ++k) tmp += xval[i][k] * xval[j][k];
//      if (i == j) gsl_matrix_set(result, i, j, tmp + lambda);
//      else gsl_matrix_set(result, i, j, tmp);
//    }
//
//  gsl_permutation *p = gsl_permutation_alloc(xval.size());
//  int s;
//  gsl_linalg_LU_decomp (result, p, &s);
//  gsl_linalg_LU_invert(result, p, inverse);
//  gsl_matrix_free(result);
//  gsl_matrix *v = gsl_matrix_alloc(xval.size(), xval[0].size());  
//  for (long i = 0; i < xval.size(); ++i)
//    for (long j = 0; j < xval[0].size(); ++j){
//      double tmp = 0;
//      for (long k = 0; k < xval.size(); ++k)
//        tmp += gsl_matrix_get(inverse, i, k) * xval[k][j];
//      gsl_matrix_set(v, i, j, tmp);
//    }
//  gsl_matrix_free(inverse);
//  gsl_permutation_free(p);
//        pred.clear();
//  pred.resize(xval.size(), 0);
//  stat.negw = 0;
//  for (long i = 0; i < xval.size(); ++i){
//    for (long k = 0; k < yval.size(); ++k) 
//      pred[i] = pred[i] + gsl_matrix_get(v, i, k) * yval[k];
//  }
  return 0;
}


double getexonweight(int exonlen, int readlen){
  double multiplier=2.0;
  double ret=exonlen*1.0/multiplier/readlen;
  if(ret>1.0)return 1.0;
  return ret;
}

/*
  Predict expression levels of sub-instance 
  Notice that res must be binary matrix
  The result exp is a Nisoforms*Nsamples matrix
*/
int predsubinst(MInst&inst, mi& res, md& exp){

 
  int nsegs=inst.getNSegs();
  /*
  vi mintersect(nsegs);
  for(int i=0;i<res.size();i++)
    vecunion(mintersect,res[i]);
  int nselexon=0;for(int i=0;i<mintersect.size();i++)nselexon+=mintersect[i];
  //calculate length
  vd exonlen(nsegs,0.0);
  for(int i=0;i<nsegs;i++) exonlen[i]=inst.getSegLen(i);
  //remove unselected columns
  //notice that xval is nselexon*nisoform , and yval is nselexon * 1
  md xval(nselexon,vd(res.size(),0)); 
  vd yval(nselexon,0);
  exp.assign(res.size(),vd(inst.getNSamples(),0));
  for(int i=0;i<inst.getNSamples();i++){
    int crdlen=inst.getReadLen(i); if(crdlen<=0)crdlen=100;
    //set up xval and yval
    int index=0;
    for(int j=0;j<mintersect.size();j++){
      double weight=getexonweight(exonlen[j],crdlen);
      if(mintersect[j]>0){
        for(int k=0;k<res.size();k++){
          xval[index][k]=res[k][j]*weight;
        }
        yval[index]=inst.segcount[j][i]*1.0/exonlen[j]*weight;  //this is the expression level
        index++;
      }
    }//end for j
    //predict
    vd pred;
    gsllassoconstraint(inst, xval,  yval,  pred);

    cout<<"Expression levels of isoforms of sample "<<i<<":";
    if(pred.size()!=res.size())cout<<"Error: incorrect pred size ("<<pred.size()<<") vs ("<<res.size()<<")\n";
    for(int j=0;j<res.size();j++){
      exp[j][i]=pred[j];
      cout<<pred[j]<<" ";
    }
    cout<<endl;
  }//end for i
    */


  //implemented by EM
  
  //setting up NewInstance
  //Need to fill out SGCounts and SGTypes, exonLen, readLen
  NewInstance newins;
  newins.nExons=nsegs;
  newins.SGTypes.resize(inst.type.size(), vector<long>(nsegs,0));
  cout<<"SGTypes size:"<<inst.type.size()<<endl;
  for(int i=0;i<inst.type.size();i++)
  {
    for(int j=0;j<inst.type[i].size();j++)
    {
      int cseg=inst.type[i][j];
      newins.SGTypes[i][cseg]=1;
    }
  }

  for(int i=0;i<nsegs;i++) newins.exonLen.push_back(inst.getSegLen(i));

  //converting res (mi) to alliso (ml);
  ml alliso(res.size());
  for(int i=0;i<res.size();i++)
  {
    cout<<"ALLISO:"<<endl;
    for(int j=0;j<res[i].size();j++) 
    {
      int ksg=res[i][j];
      alliso[i].push_back(ksg);
      cout<<ksg<<" ";
    }
    cout<<endl;
  }
  //setting up EM parameters
  vector<string> varpara;
  varpara.push_back("--no-filter");
  varpara.push_back("--verbose");

  //setting up return values of EM
  ml pred;
  vd predexplv;
  STAT stat;
  exp.assign(res.size(),vd(inst.getNSamples(),0));
  
  for(int i=0;i<inst.getNSamples();i++){
    //fill out SGCounts and SGTypes, exonLen, readLen
    newins.SGCounts.resize(inst.typecount.size());
    for(int j=0;j<inst.typecount.size();j++) newins.SGCounts[j]=inst.typecount[j][i];
    //cout<<"SGCOUNTS:\n";for(int j=0;j<newins.SGCounts.size();j++) cout<<newins.SGCounts[j]<<" "; cout<<endl;
    newins.readLen=inst.readlen[i];
    
    
    //calling the EM function
    emselect(newins,alliso,varpara,pred,predexplv,stat);

    //check the expression
    cout<<"Expression levels of isoforms of sample "<<i<<":";
    //for(int j=0;j<predexplv.size();j++) cout<<predexplv[j]<<" ";cout<<endl;
    if(predexplv.size()!=res.size())cout<<"Error: incorrect pred size ("<<predexplv.size()<<") vs ("<<res.size()<<")\n";
    for(int j=0;j<res.size();j++){
      exp[j][i]=predexplv[j];
      //cout<<predexplv[j]<<" ";
    }
    //cout<<endl;

  }
  return 0;
}
/*REPLACED BY predsubinst
  Predict expression levels of sub-instance 
  Notice that res must be binary matrix
  The result exp is a Nisoforms*Nsamples matrix
*/
int predsubinst0(MInst&inst, mi& res, md& exp){

 /*
  int nsegs=inst.getNSegs();
  vi mintersect(nsegs);
  for(int i=0;i<res.size();i++)
    vecunion(mintersect,res[i]);
  int nselexon=0;for(int i=0;i<mintersect.size();i++)nselexon+=mintersect[i];
  //calculate length
  vd exonlen(nsegs,0.0);
  for(int i=0;i<nsegs;i++) exonlen[i]=inst.getSegLen(i);
  //remove unselected columns
  //notice that xval is nselexon*nisoform , and yval is nselexon * 1
  md xval(nselexon,vd(res.size(),0)); 
  vd yval(nselexon,0);
  exp.assign(res.size(),vd(inst.getNSamples(),0));
  for(int i=0;i<inst.getNSamples();i++){
    int crdlen=inst.getReadLen(i); if(crdlen<=0)crdlen=100;
    //set up xval and yval
    int index=0;
    for(int j=0;j<mintersect.size();j++){
      double weight=getexonweight(exonlen[j],crdlen);
      if(mintersect[j]>0){
        for(int k=0;k<res.size();k++){
          xval[index][k]=res[k][j]*weight;
        }
        yval[index]=inst.segcount[j][i]*1.0/exonlen[j]*weight;  //this is the expression level
        index++;
      }
    }//end for j
    //predict
    vd pred;
    gsllassoconstraint(inst, xval,  yval,  pred);
    cout<<"Expression levels of isoforms of sample "<<i<<":";
    if(pred.size()!=res.size())cout<<"Error: incorrect pred size ("<<pred.size()<<") vs ("<<res.size()<<")\n";
    for(int j=0;j<res.size();j++){
      exp[j][i]=pred[j];
      cout<<pred[j]<<" ";
    }
    cout<<endl;
  }//end for i
  */
  return 0;
}

/* prediction expression level based on the assembly results in res */
int calcexplv(MInst& inst, mi& res, md& exp){
  //separate the predictions into different groups
  //first, create a fingerprint map
  int nsegs=inst.getNSegs();
  mi mbin(res.size(),vi(nsegs,0));
  for(int i=0;i<res.size();i++)
    for(int j=0;j<res[i].size();j++) mbin[i][res[i][j]]=1;
  
  mi munionindex;
  //group the results
  groupres(mbin,munionindex);
  
  //predict the sub instance separately
  exp=md(res.size());
  for(int i=0;i<munionindex.size();i++){
    cout<<"Group:";for(int j=0;j<munionindex[i].size();j++) cout<<munionindex[i][j]<<" ";cout<<endl;
    for(int j=0;j<munionindex[i].size();j++) {
      int t=munionindex[i][j];
      for(int k=0;k<res[t].size();k++)
        cout<<res[t][k]<<" ";
      cout<<endl;
    }
    mi sres;md sexp;
    for(int j=0;j<munionindex[i].size();j++)
      sres.push_back(mbin[munionindex[i][j]]);
    predsubinst(inst,sres,sexp);
    //write back to results
    for(int j=0;j<munionindex[i].size();j++) 
      exp[munionindex[i][j]]=sexp[j];
  }
  return 0;
}


