#include <iostream>
#include <fstream>
#include <sstream>
#include "minst.h"
#include "menumiso.h"
#include "mfileio.h"
#include "mflow.h"
#include "mpredexp.h"

using namespace std;

/* Do we want to read some specific instance? */
string objid="";
string outfile="";
string outexpfile="";

/* For pthread option, only check instances with id%pthread_all==pthread_this*/
int pthread_all=0;
int pthread_this=0;


EnumParameter THISPARA;

void printhelp(){
  cout<<"Usage: predminst {OPTIONS} <instance file>\n\n";
  cout<<"Options:\n";
  cout<<"\t-h\t\t\tPrint help information\n";
  cout<<"\t-i <ID>\t\t\tPredict only the instance with specified ID.\n"; 
  cout<<"\t-o <file>\t\tThe prefix of the output file.\n";
  cout<<"\t-p x,y\t\t\tOnly predict instances with ID pattern x,y. For example, if (x,y)=(3,4), the program will only predict instances with ID 3,7,11,15,...\n"; 
  cout<<"\t--min-frac <0.0-1.0>\tOnly predict isoforms with expression greater than this fraction of the most abundant isoform. Default 0.1.\n";
  cout<<"\t--rd-alpha <0.0-1.0>\tThe weights for read supporting variables. Default 0.1 \n";
  cout<<"\t--assemble-by-sample\tAssemble the transcripts sample-by-sample. This will lead to higher sensitivity but lower precision.\n";
  
  cout<<"\t--no-correlation\tDo not perform segment correlation.\n";

}

/* Parse the option 
Return -1 if an error occurs.
*/
int parseopt(int argc,char* argv[]){
  int ii=0;
  while(ii<argc){
    if(string(argv[ii])=="-h"){
      printhelp();
      return -1;
    }
    if(string(argv[ii])=="-e" && ii<argc-1){
      MInst::readEJ(argv[ii+1]);
      ii++;
    }
    if(string(argv[ii])=="-i" && ii<argc-1){
      objid=string(argv[ii+1]);
      ii++;
    }
    if(string(argv[ii])=="-o" && ii<argc-1){
      outfile=string(argv[ii+1]);
      outexpfile=outfile+".explv";
      ii++;
    }
    if(string(argv[ii])=="-p" && ii<argc-1){
      string tsx(argv[ii+1]);
      stringstream ss(tsx);
      char c;
      ss>>pthread_this>>c>>pthread_all;
      cout<<"Pthread pattern: "<<pthread_this<<"/"<<pthread_all<<endl;
    }
    if(string(argv[ii])=="--min-frac" && ii<argc-1){
      string tsx(argv[ii+1]);
      stringstream ss(tsx);
      ss>>THISPARA.minjuncfrac;
      cout<<"min-frac:"<<THISPARA.minjuncfrac<<endl;
    }
    if(string(argv[ii])=="--rd-alpha" && ii<argc-1){
      string tsx(argv[ii+1]);
      stringstream ss(tsx);
      ss>>THISPARA.RDSUPPORT_ALPHA;
      cout<<"rd-alpha:"<<THISPARA.RDSUPPORT_ALPHA<<endl;
    }
    if(string(argv[ii])=="--assemble-by-sample"){
      THISPARA.assemble_by_sample=true;
    }
    if(string(argv[ii])=="--no-correlation"){
      THISPARA.DO_CORRELATION=false;
    }
    ii++;
  }
  return 0;
}

int main(int argc, char* argv[]){
  //set up default parameters
  THISPARA.minintronretentioncvgfrac=0.2;
  //para.minjuncfrac=0.1;para.minadjjuncfrac=1.4;
  THISPARA.debug=true;

  if(argc==1){printhelp();return 0;}
  if(parseopt(argc,argv)==-1) return -1;
  ifstream ifs(argv[argc-1]);
  if(!ifs.is_open()){
    cerr<<"Error opening file "<<argv[argc-1]<<endl;
    return -1;
  }
  if(outfile==""){
    outfile=string(argv[argc-1]);
    if(objid!="") outfile+="_"+objid;
    outfile+=".pred";
    outexpfile=outfile+".explv";
  }
  ofstream ofs(outfile.c_str());
  if(!ofs.is_open()){
    cerr<<"Error opening file "<<outfile<<endl;
    return -1;
  }
  ofstream oexpfs(outexpfile.c_str());
  if(!oexpfs.is_open()){
    cerr<<"Error opening file "<<outexpfile<<endl;
    return -1;
  }
  while(true){
    MInst mis;
    mis.clear();
    if(ifs.eof()){
      break;
    }
    if(mis.readFromStream(ifs,objid,pthread_this,pthread_all)==-1)break;
    if(mis.check()==-1){
      cerr<<"Error: check failed.\n";
      return -1;
    }
    
    mis.converttype();
    cout<<"Instance "<<mis.id<<"..."<<endl;
    mi res;
    vd exp;
    md explv;
    vi direction;
    //the old enumeration 
    //enumandcalexp(mis,  res, explv);
  
    //the new methods, using flow algorithm to calculate expression levels
    flowandcalexp(mis, res,  explv,direction,THISPARA);
  
    //predict expression level
    cout<<"# of predicted isoforms: "<<res.size()<<endl;
    // write to file
    predToStream(ofs,oexpfs,mis, res, explv,direction);
    if(objid!="")break;
  }
  oexpfs.close();
  ofs.close();
  ifs.close();
  return 0;
}
